﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectSound;

namespace Bvews.DirectSound
{
    public class DirectSoundManager : IDisposable
    {
        Device device;
        SecondaryBuffer buffer;
        int bufferBytes;
        int nextWrite;
        System.Timers.Timer timer;
        bool ownsDevice;
        Func<int, byte[]> createBufferData;

        ulong count;
        double phase;
        
        public DirectSoundManager(Control owner, WaveFormat format) :
            this(owner, null, format) { }

        public DirectSoundManager(Control owner, Device device, WaveFormat format)
        {
            this.device = device;
            if (this.device == null)
            {
                this.device = new Device();
                this.device.SetCooperativeLevel(
                    owner, CooperativeLevel.Normal);
                ownsDevice = true;
            }

            //WaveFormat format = new WaveFormat();
            //format.SamplesPerSecond = 48000;
            //format.Channels = 1;
            //format.BitsPerSample = 16;
            //format.BlockAlign = (short)(format.Channels * (format.BitsPerSample / (short)8));
            //format.AverageBytesPerSecond = format.BlockAlign * format.SamplesPerSecond;

            BufferDescription desc = new BufferDescription(format);
            desc.BufferBytes = format.AverageBytesPerSecond / 10;
            desc.ControlVolume = true;
            desc.GlobalFocus = true;

            buffer = new SecondaryBuffer(desc, this.device);
            bufferBytes = buffer.Caps.BufferBytes;

            timer = new System.Timers.Timer(BytesToMs(bufferBytes / 6));
            timer.Enabled = true;
            timer.Elapsed += timer_Elapsed;

            createBufferData = (length) =>
            {
                return new byte[length];
            };
        }

        ~DirectSoundManager()
        {
            Dispose();
        }

        public Func<int, byte[]> CreateBufferData
        {
            get { return createBufferData; }
            set { createBufferData = value; }
        }

        int MaxLatencyMs
        {
            get { return 300; }
        }

        public static WaveFormat GetWaveFormat(int samplesPerSecond, short channels, short bitsPerSample)
        {
            WaveFormat format = new WaveFormat();
            format.SamplesPerSecond = samplesPerSecond;
            format.Channels = channels;
            format.BitsPerSample = bitsPerSample;
            format.BlockAlign = (short)(format.Channels * (format.BitsPerSample / (short)8));
            format.AverageBytesPerSecond = format.BlockAlign * format.SamplesPerSecond;
            return format;
        }

        public void Play()
        {
            Stop();

            //m_PullStream = new PullStream(pullAudio);

            buffer.SetCurrentPosition(0);
            nextWrite = 0;
            Feed(bufferBytes);
            timer.Enabled = true;
            buffer.Play(0, BufferPlayFlags.Looping);
        }

        public void Stop()
        {
            if (timer != null)
            {
                timer.Enabled = false;
            }

            if (timer != null)
            {
                if (!buffer.Disposed)
                {
                    buffer.Stop();
                }
            }
        }

        public void Dispose()
        {
            Stop();
            if (timer != null)
            {
                timer.Dispose();
                timer = null;
            }
            if (buffer != null)
            {
                buffer.Dispose();
                buffer = null;
            }
            if (ownsDevice && device != null)
            {
                device.Dispose();
                device = null;
            }
            GC.SuppressFinalize(this);
        }

        private void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            Feed(GetPlayedSize());
        }

        private int GetPlayedSize()
        {
            int pos = buffer.PlayPosition;
            return
               pos < nextWrite ?
               pos + bufferBytes - nextWrite
               : pos - nextWrite;
        }

        private void Feed(int bytes)
        {
            // 遅延を数ミリ秒に制限
            int tocopy = Math.Min(bytes, MsToBytes(MaxLatencyMs));

            if (tocopy > 0)
            {
                // バッファを復元
                if (buffer.Status.BufferLost)
                {
                    buffer.Restore();
                }

                // データをバッファにコピー
                //using (var pullStream = CreateWaveDataStream(tocopy))
                {
                    //var pullStream = CreateWaveDataStream(tocopy);
                    //buffer.Write(nextWrite, pullStream, tocopy, LockFlag.None);

                    var waveData = new byte[tocopy];
                    waveData = createBufferData(tocopy);

                    ////  8bit: length
                    //// 16bit: length / 2
                    //for (int i = 0; i < tocopy / 2; i++)
                    //{
                    //    byte[] byteArray = BitConverter.GetBytes((ushort)(2048 * Math.Sin(phase)));
                    //    waveData[i * 2 + 0] = byteArray[0];
                    //    waveData[i * 2 + 1] = byteArray[1];
                    //    count++;

                    //    phase += 2 * Math.PI * frequency / 48000.0;
                    //    if (phase > 2 * Math.PI)
                    //    {
                    //        phase = phase % (2 * Math.PI);
                    //    }
                    //}
                    if (buffer != null)
                    {
                        buffer.Write(nextWrite, waveData, LockFlag.None);
                    }
                }

                nextWrite += tocopy;
                if (nextWrite >= bufferBytes) nextWrite -= bufferBytes;
            }
        }
        
        private int BytesToMs(int bytes)
        {
            return bytes * 1000 / buffer.Format.AverageBytesPerSecond;
        }

        private int MsToBytes(int ms)
        {
            int bytes = ms * buffer.Format.AverageBytesPerSecond / 1000;
            bytes -= bytes % buffer.Format.BlockAlign;
            return bytes;
        }
    }
}
