﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading.Tasks;
using System.IO;

namespace Bvews.DirectSound
{
    class StreamingPlayer
    {
        Timer timer = new Timer();
        DirectSoundManager manager;
        System.Windows.Forms.Control owner;

        DateTime playStartTime;
        DateTime startPositionTime;
        TimeSpan duration;
        int currentPosition;
        long wavDurationTicks;
        int concreteIncrement = 1;
        int increment = 1;

        int cursorIncrementSign = 1;
        long cursorPositionTicks = 0;
        DateTime prevDateTime;

        bool endless = true;
        bool reverseLoop = true;

        int sampleRate = 48000;
        int channels = 1;
        const int bitsPerSample = 16;
        
        public StreamingPlayer(System.Windows.Forms.Control owner)
        {
            this.owner = owner;
        }

        public Func<int, float[][]> CreateBufferData;

        public event EventHandler Tick;
        public event EventHandler Stopped;

        public int SampleRate
        {
            get { return sampleRate; }
            set { sampleRate = value; }
        }

        public int Channels
        {
            get { return channels; }
            set { channels = value; }
        }

        public bool Endless
        {
            get { return endless; }
            set { endless = value; }
        }

        public bool ReverseLoop
        {
            get { return reverseLoop; }
            set { reverseLoop = value; }
        }

        public System.Windows.Forms.Control Owner
        {
            get { return owner; }
        }

        public void Play()
        {
            Play(null);
        }

        public void Play(double? rate)
        {
            timer.Enabled = true;

            playStartTime = DateTime.Now;
            startPositionTime = new DateTime((long)(10000000L * currentPosition / (double)sampleRate));
            cursorPositionTicks = startPositionTime.Ticks;
            prevDateTime = DateTime.Now;
            
            if (manager == null)
            {
                Microsoft.DirectX.DirectSound.WaveFormat format = DirectSoundManager.GetWaveFormat(sampleRate, (short)channels, (short)bitsPerSample);
                manager = new DirectSoundManager(owner, format);
            }

            manager.CreateBufferData = (tocopy) =>
            {         
                byte[] bufferData = new byte[tocopy];
                if(CreateBufferData == null)
                {
                    return bufferData;
                }
                
                var bytesPerSample = bitsPerSample / 8;
                var blockSize = bytesPerSample * channels;
                int length = tocopy / blockSize;

                var bufs = CreateBufferData(length);
                if(bufs == null)
                {
                    return bufferData;
                }

                for (int i = 0; i < length; i++)
                {
                    for (int j = 0; j < channels; j++)
                    {
                        var bytes = BitConverter.GetBytes((short)(bufs[j][i] * 32768));
                        Array.Copy(bytes, 0, bufferData, i * blockSize + j * bytesPerSample, 2);
                    }
                }
                return bufferData;
            };

            manager.Play();
        }

        public void Stop()
        {
            //drawXorLine.DeleteVerticalLine();
            if (Stopped != null)
            {
                Stopped(this, EventArgs.Empty);
            }

            timer.Enabled = false;
            if (manager != null)
            {
                manager.Stop();
                manager.Dispose();
                manager = null;
            }
        }
        
        private void timer_Tick(object sender, EventArgs e)
        {
            if (manager == null)
            {
                return;
            }

            var timeElapsed = DateTime.Now - prevDateTime;
            prevDateTime = DateTime.Now;

            bool stopPlaying = false;
            cursorPositionTicks += timeElapsed.Ticks * cursorIncrementSign;
            
            // Endless loop
            if (endless)
            {
                if (reverseLoop)
                {
                    if (cursorPositionTicks < 0)
                    {
                        cursorIncrementSign = 1;
                        cursorPositionTicks -= cursorPositionTicks % wavDurationTicks;
                    }
                    else if (cursorPositionTicks >= wavDurationTicks)
                    {
                        cursorIncrementSign = -1;
                        cursorPositionTicks -= cursorPositionTicks % wavDurationTicks;
                    }
                }
                else
                {
                    if (cursorPositionTicks < 0)
                    {
                        cursorPositionTicks = cursorPositionTicks % wavDurationTicks + wavDurationTicks;
                    }
                    else if (cursorPositionTicks >= wavDurationTicks)
                    {
                        cursorPositionTicks = cursorPositionTicks % wavDurationTicks;
                    }
                }
            }
            else
            {
                if (cursorPositionTicks < 0)
                {
                    cursorPositionTicks = 0;
                    stopPlaying = true;
                }
                else if (cursorPositionTicks > wavDurationTicks)
                {
                    cursorPositionTicks = wavDurationTicks;
                    stopPlaying = true;
                }
            }

            //int currentPosX = (int)(cursorPositionTicks / (double)wavDurationTicks * pictureBox.Width);
            //drawXorLine.DrawVerticalLine(currentPosX);
            //timeLabel.Text = new DateTime(ticks).ToString("HH:mm:ss.fff");


            if (!endless && (duration < timeElapsed || stopPlaying))
            {
                Stop();
            }

            if (Tick != null)
            {
                Tick(this, EventArgs.Empty);
            }
        }
    }
}
