﻿#nullable enable

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Bvews.Common;
using Bvews.Wave;
using Bvews.DirectSound;
using System.Windows.Forms.VisualStyles;
using BveTool;

namespace AudioLoopMaker
{
    public partial class FormMain : Form
    {
        WaveData? waveData;
        WaveStream? waveStream;
        WaveStreamOrdinaryLoop? waveStreamOrdinaryLoop;
        WaveStreamAlternateLoop? waveStreamAlternateLoop;
        StreamingPlayer player;
        string inFilePath = string.Empty;
        bool isListening = false;

        public FormMain(string[] args)
        {
            InitializeComponent();

            comboBoxZoom.SelectedIndex = 0;

            player = new StreamingPlayer(this);
            player.CreateBufferData = length =>
            {
                if (waveStream != null)
                {
                    pictureBoxWaveSummary.CursorPosition = waveStream.BufferPosition;
                    return waveStream.Read(length);
                }
                else
                {
                    return null;
                }
            };

            if (LanguageSettingManager.IsJapanese)
            {
                japaneseToolStripMenuItem.Checked = true;
                englishToolStripMenuItem.Checked = false;
            }

            Size = Settings.Instance.Preferences.WindowSize;

            pictureBoxWaveLeft.WaveViewOptions = Settings.Instance.WaveViewOptions;
            pictureBoxWaveRight.WaveViewOptions = Settings.Instance.WaveViewOptions;
            pictureBoxWaveSummary.WaveViewOptions = Settings.Instance.WaveViewOptions;

            checkBoxAddSuffix.Checked = Settings.Instance.Preferences.AddSuffix;
            radioButtonOrdinary.Checked = Settings.Instance.Preferences.Loop == Loop.Ordinary;
            radioButtonAlternate.Checked = Settings.Instance.Preferences.Loop == Loop.Alternate;

#if DEBUG
            try { LoadFile("audio.wav"); }
            catch (Exception) { }
#endif

            if (args != null && args.Length > 0)
            {
                LoadFile(args[0]);
            }
        }

        private void LoadFile(string path)
        {
            try
            {
                if (!path.ToLower().EndsWith(".wav"))
                {
                    MessageBox.Show(Properties.Resources.NotWavFile, Properties.Resources.Warning, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    return;
                }

                Pause();

                var reader = new WaveDataReader(path);
                var waveDataTemp = reader.Read();
                if (waveDataTemp.Buffers.Min(b => b.Length) == 0)
                {
                    MessageBox.Show(Properties.Resources.HasNoWaveform, Properties.Resources.Warning, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    return;
                }
                waveData = waveDataTemp;
                player.SampleRate = waveData.WaveFormat.SampleRate;
                player.Channels = waveData.WaveFormat.Channels;

                var buffer = AudioBufferUtils.FlattenToMono(waveData.Buffers);
                pictureBoxWaveSummary.Buffer = buffer;
                pictureBoxWaveLeft.Buffer = waveData.Buffers[0];
                pictureBoxWaveRight.Buffer = waveData.Buffers.Length == 1 ?
                    waveData.Buffers[0] : waveData.Buffers[1];
                nudPosition.Maximum = buffer.Length;

                var maxAmplitude = (float)AudioBufferUtils.GetMaxAmplitude(waveData.Buffers);
                pictureBoxWaveLeft.MaxAmplitude = maxAmplitude;
                pictureBoxWaveRight.MaxAmplitude = maxAmplitude;

                waveStreamOrdinaryLoop = new WaveStreamOrdinaryLoop(waveData.Buffers);
                waveStreamAlternateLoop = new WaveStreamAlternateLoop(waveData.Buffers);

                if (radioButtonOrdinary.Checked)
                {
                    waveStream = waveStreamOrdinaryLoop;
                }
                else if (radioButtonAlternate.Checked)
                {
                    waveStream = waveStreamAlternateLoop;
                }

                inFilePath = path;
                saveFileDialog.FileName = string.Empty;
                radioButtonStart.Checked = true;
                nudPosition.Value = 0M;

                UpdateView();
            }
            catch
            {
#if DEBUG
                throw;
#else
                MessageBox.Show(Properties.Resources.FailedToLoad, Properties.Resources.Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
#endif           
            }
        }

        private void SaveFile(string path)
        {
            if (waveData == null || waveStream == null)
            {
                return;
            }

            try
            {
                var writer = new WaveDataWriter();
                writer.Write(path, waveData.WaveInfo, waveStream.Flush());
            }
            catch (Exception)
            {
#if DEBUG
                throw;
#else
                MessageBox.Show(Properties.Resources.FailedToSave, Properties.Resources.Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
#endif    
            }
        }

        private void Listen()
        {
            player.Play();
            isListening = true;
            buttonListen.Text = Properties.Resources.Pause;
        }

        private void Pause()
        {
            player.Stop();
            isListening = false;
            buttonListen.Text = Properties.Resources.Listen;
        }

        private void Open()
        {

        }

        private void Save()
        {
            if (waveData == null || waveStream == null)
            {
                return;
            }

            SaveFile(inFilePath);
        }

        private void SaveAs()
        {
            if (waveData == null || waveStream == null)
            {
                return;
            }

            try
            {
                if (File.Exists(saveFileDialog.FileName))
                {
                    var fileName = saveFileDialog.FileName;
                    saveFileDialog.FileName = Path.GetFileNameWithoutExtension(fileName);
                    saveFileDialog.InitialDirectory = Path.GetDirectoryName(fileName);
                }
                else
                {
                    var fileNameWithoutExtension = Path.GetFileNameWithoutExtension(inFilePath);
                    fileNameWithoutExtension += checkBoxAddSuffix.Checked ? "_l" : string.Empty;
                    saveFileDialog.FileName = fileNameWithoutExtension;
                    saveFileDialog.InitialDirectory = Path.GetDirectoryName(inFilePath);
                }
            }
            catch { }

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                SaveFile(saveFileDialog.FileName);
            }
        }

        private void UpdateView()
        {
            if (File.Exists(inFilePath) && Settings.Instance.Preferences.ShowFileNameInTitle)
            {
                Text = $"{Path.GetFileName(inFilePath)} - Audio Loop Maker";
            }
            else
            {
                Text = "Audio Loop Maker";
            }

            pictureBoxWaveLeft.Invalidate();
            pictureBoxWaveRight.Invalidate();
            pictureBoxWaveSummary.Invalidate();
        }

        private void pictureBoxWaveSummary_SectionStartChanged(object sender, EventArgs e)
        {
            radioButtonStart.Checked = true;
            if (sender is PictureBoxWaveSummary pictureBox)
            {
                nudPosition.Value = pictureBox.StartPosition;
                if (waveStream != null)
                {
                    waveStream.Start = pictureBox.StartPosition;
                }
            }
        }

        private void pictureBoxWaveSummary_SectionEndChanged(object sender, EventArgs e)
        {
            radioButtonEnd.Checked = true;
            if (sender is PictureBoxWaveSummary pictureBox)
            {
                nudPosition.Value = pictureBox.EndPosition;
                if (waveStream != null)
                {
                    waveStream.End = pictureBox.EndPosition;
                }
            }
        }

        private void nudPosition_ValueChanged(object sender, EventArgs e)
        {
            if (sender is NumericUpDown nud)
            {
                var value = (int)nud.Value;
                if (radioButtonStart.Checked)
                {
                    pictureBoxWaveSummary.StartPosition = value;
                    pictureBoxWaveLeft.StartPosition = value;
                    pictureBoxWaveRight.StartPosition = value;
                    if (waveStream != null)
                    {
                        waveStream.Start = value;
                    }
                }
                else if (radioButtonEnd.Checked)
                {
                    pictureBoxWaveSummary.EndPosition = value;
                    pictureBoxWaveLeft.EndPosition = value;
                    pictureBoxWaveRight.EndPosition = value;
                    if (waveStream != null)
                    {
                        waveStream.End = value;
                    }
                }
            }
        }

        private void radioButtonStart_CheckedChanged(object sender, EventArgs e)
        {
            nudPosition.Value = (decimal)pictureBoxWaveSummary.StartPosition;
            pictureBoxWaveLeft.Position = Position.Start;
            pictureBoxWaveRight.Position = Position.Start;
        }

        private void radioButtonEnd_CheckedChanged(object sender, EventArgs e)
        {
            nudPosition.Value = (decimal)pictureBoxWaveSummary.EndPosition;
            pictureBoxWaveLeft.Position = Position.End;
            pictureBoxWaveRight.Position = Position.End;
        }

        private void buttonListen_Click(object sender, EventArgs e)
        {
            ToggleListenState();
        }

        private void ToggleListenState()
        {
            if (isListening)
            {
                Pause();
            }
            else
            {
                Listen();
            }
        }

        private void radioButtonLoopType_CheckedChanged(object sender, EventArgs e)
        {
            Pause();

            if (radioButtonOrdinary.Checked)
            {
                if(waveStreamOrdinaryLoop != null && waveStreamAlternateLoop != null)
                {
                    waveStreamOrdinaryLoop.Start = waveStreamAlternateLoop.Start;
                    waveStreamOrdinaryLoop.End = waveStreamAlternateLoop.End;
                }
                waveStream = waveStreamOrdinaryLoop;
                pictureBoxWaveLeft.Loop = Loop.Ordinary;
                pictureBoxWaveRight.Loop = Loop.Ordinary;
            }
            else if (radioButtonAlternate.Checked)
            {
                if (waveStreamOrdinaryLoop != null && waveStreamAlternateLoop != null)
                {
                    waveStreamAlternateLoop.Start = waveStreamOrdinaryLoop.Start;
                    waveStreamAlternateLoop.End = waveStreamOrdinaryLoop.End;
                }
                waveStream = waveStreamAlternateLoop;
                pictureBoxWaveLeft.Loop = Loop.Alternate;
                pictureBoxWaveRight.Loop = Loop.Alternate;
            }

            if (waveStream != null)
            {
                pictureBoxWaveSummary.StartPosition = waveStream.Start;
                pictureBoxWaveSummary.EndPosition = waveStream.End;

                if (radioButtonStart.Checked)
                {
                    nudPosition.Value = nudPosition.Value = (decimal)waveStream.Start;
                    pictureBoxWaveLeft.Position = Position.Start;
                    pictureBoxWaveRight.Position = Position.Start;
                }
                else if (radioButtonEnd.Checked)
                {
                    nudPosition.Value = nudPosition.Value = (decimal)waveStream.End;
                    pictureBoxWaveLeft.Position = Position.End;
                    pictureBoxWaveRight.Position = Position.End;
                }
            }
        }


        private void FormMain_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] drags = (string[])e.Data.GetData(DataFormats.FileDrop);
                e.Effect = DragDropEffects.Copy;
            }

        }

        private void FormMain_DragDrop(object sender, DragEventArgs e)
        {
            var paths = (string[])e.Data.GetData(DataFormats.FileDrop);
            LoadFile(paths[0]);
        }

        private void buttonSaveAs_Click(object sender, EventArgs e)
        {
            SaveAs();
        }

        private void comboBoxZoom_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (sender is ComboBox comboBox)
            {
                pictureBoxWaveLeft.ZoomOutExponent = comboBox.SelectedIndex;
                pictureBoxWaveRight.ZoomOutExponent = comboBox.SelectedIndex;
            }
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if(openFileDialog.ShowDialog() == DialogResult.OK)
            {
                LoadFile(openFileDialog.FileName);
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveAs();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            using var form = new PreferencesForm();
            if(form.ShowDialog() == DialogResult.OK)
            {
                UpdateView();
            }
        }

        private void showHelpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            HelpManager.ShowHelp("https://bvews.jpn.org/materials/audio-loop-maker/");
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var license = "MIT License";
            try { license = File.ReadAllText(Path.Combine(Application.StartupPath, "LICENSE")); }
            catch { }

            using var form = new AboutForm();
            form.ShowVersionInfoWithLicense(this, license, null);
        }

        private void japaneseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!japaneseToolStripMenuItem.Checked)
            {
                LanguageSettingManager.Save("ja");
                Settings.Instance.Language = "ja";
                if (MessageBox.Show(Bvews.Common.Resources.MessegeReLaunchApp, Bvews.Common.Resources.TextConfirm, MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    LanguageSettingManager.Relaunch = true;
                    Pause();
                    Close();
                }
            }
            japaneseToolStripMenuItem.Checked = true;
            englishToolStripMenuItem.Checked = false;
        }

        private void englishToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!englishToolStripMenuItem.Checked)
            {
                LanguageSettingManager.Save("en");
                Settings.Instance.Language = "en";
                if (MessageBox.Show(Bvews.Common.Resources.MessegeReLaunchApp, Bvews.Common.Resources.TextConfirm, MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    LanguageSettingManager.Relaunch = true;
                    Pause();
                    Close();
                }
            }
            japaneseToolStripMenuItem.Checked = false;
            englishToolStripMenuItem.Checked = true;
        }

        private void FormMain_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Space)
            {
                ToggleListenState();
                e.Handled = true; 
            }
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            Settings.Instance.Preferences.WindowSize = Size;
            Settings.Instance.Preferences.AddSuffix = checkBoxAddSuffix.Checked;
            Settings.Instance.Preferences.Loop = radioButtonOrdinary.Checked ? Loop.Ordinary : Loop.Alternate;

            Settings.Save();
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            var notifyUpdate = Settings.Instance.Preferences.NotifyUpdate;
            if (!notifyUpdate)
            {
                return;
            }

            var form = new Bvews.Common.UpdateChecker.UpdateNotificationForm();
            form.DoNotShowAgain = !notifyUpdate;
            form.OnUpdateInfoLoad += (s, ev) =>
            {
                if (!form.IsLatest && form.ShowDialog() == DialogResult.OK)
                {
                    Settings.Instance.Preferences.NotifyUpdate = !form.DoNotShowAgain;
                }
            };
            form.LoadUpdateInfoAsync("https://bvews.jpn.org/materials/audio-loop-maker/");
        }
    }
}
