﻿#nullable enable

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace AudioLoopMaker
{
    class PictureBoxWaveDetail : PictureBox
    {
        WaveViewOptions options = new WaveViewOptions();

        Point? sectionBeginMoveBeginLocation = null;
        int startX;
        int startXdelta;

        Point? sectionEndMoveBeginLocation = null;
        int endX;
        int endXdelta;

        int cursorX;

        float maxAmplitude;
        Position position;
        Loop loop;
        int zoomOutExponent;

        float[] buffer = new float[0];
        int startPosition;
        int endPosition;
        int cursorPosition;

        PointF[] waveformPoints = new PointF[0];

        public event EventHandler? SectionStartChanged;
        public event EventHandler? SectionEndChanged;

        internal WaveViewOptions WaveViewOptions { get => options; set => options = value; }

        public float[] Buffer
        {
            get => buffer;
            set
            {
                buffer = value;
                startPosition = 0;
                endPosition = buffer.Length;
                waveformPoints = ComputeWaveformPoints().ToArray();
                Invalidate();
            }
        }

        public int StartPosition
        {
            get => startPosition;
            set
            {
                startPosition = value;
                startPosition = startPosition > 0 ? startPosition : 0;
                startPosition = startPosition < endPosition ? startPosition : endPosition;
                if (buffer.Length > 0)
                {
                    startX = startPosition * Width / buffer.Length;
                }
                Position = Position.Start;
                Invalidate();
            }
        }

        public int EndPosition
        {
            get => endPosition;
            set
            {
                endPosition = value;
                endPosition = endPosition > 0 ? endPosition : 0;
                endPosition = endPosition < startPosition ? startPosition : endPosition;
                if (buffer.Length > 0)
                {
                    endX = endPosition * Width / buffer.Length;
                }
                Position = Position.End;
                Invalidate();
            }
        }

        internal Position Position
        {
            get => position;
            set
            {
                position = value;
                Invalidate();
            }
        }

        internal Loop Loop
        {
            get => loop;
            set
            {
                loop = value;
                Invalidate();
            }
        }

        public float MaxAmplitude
        {
            get => maxAmplitude;
            set
            {
                maxAmplitude = value;
                Invalidate();
            }
        }

        public int ZoomOutExponent
        {
            get => zoomOutExponent;
            set
            {
                zoomOutExponent = value;
                Invalidate();
            }
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            waveformPoints = ComputeWaveformPoints().ToArray();
            if (buffer.Length > 0)
            {
                startX = startPosition * Width / buffer.Length;
                endX = endPosition * Width / buffer.Length;
                cursorX = cursorPosition * Width / buffer.Length;
            }
            Invalidate();
        }

        protected override void OnPaint(PaintEventArgs pe)
        {
            base.OnPaint(pe);

            var graphics = pe.Graphics;
         
            // Waveform
            DrawWaveform(graphics);
        }

        private void DrawWaveform(Graphics g)
        {
            // Background
            g.Clear(options.Color.Background);

            // Center line
            using var centerLinePen = new Pen(options.Color.CenterLine, 1);
            g.DrawLine(centerLinePen, 0, Height / 2, Width, Height / 2);

            if (buffer == null || buffer.Length == 0)
            {
                return;
            }

            int div = ZoomOutExponent < 0 ? 1 : (int)Math.Pow(2, ZoomOutExponent);
            var scale = MaxAmplitude > 0 ? 1 / MaxAmplitude : 0;

            int center = (Width + 1) / 2;
            int height = Height;

            int begin = StartPosition;
            int end = EndPosition;
            int length = 0;

            float spliceCenter = length / 2 / div;

            int left = center - (length / 2 / div);
            int right = left + (length / div);
            
            // Splice start line
            using var sectionStartPen = new Pen(options.Color.SectionStart, 1);
            using var sectionEndPen = new Pen(options.Color.SectionEnd, 1);
            if (Loop == Loop.Ordinary)
            {
                // Ordinary Loop
                if (length == 0)
                {
                    g.DrawLine(sectionEndPen, left - 1, 0, left - 1, height);
                    g.DrawLine(sectionStartPen, right, 0, right, height);
                }
                else
                {
                    g.DrawLine(sectionStartPen, left - 1, 0, left - 1, height);
                    g.DrawLine(sectionEndPen, right, 0, right, height);
                }
            }
            else if (Loop == Loop.Alternate)
            {
                if (Position == Position.Start)
                {
                    // Start, Alternate
                    g.DrawLine(sectionStartPen, left - 1, 0, left - 1, height);
                    g.DrawLine(sectionStartPen, right, 0, right, height);
                }
                else if (Position == Position.End)
                {
                    // End, Alternate
                    g.DrawLine(sectionEndPen, left - 1, 0, left - 1, height);
                    g.DrawLine(sectionEndPen, right, 0, right, height);
                }
            }

            var waveData = buffer;

            float[] formerWaveData = new float[0];
            float[] latterWaveData = new float[0];

            if (Loop == Loop.Ordinary)
            {
                // Ordinary Loop
                formerWaveData = TrimWaveForDisplay(waveData, div, end, right, false, false);
                latterWaveData = TrimWaveForDisplay(waveData, div, begin, Width - left, true, true);
            }
            else if (Loop == Loop.Alternate)
            {
                if (Position == Position.Start)
                {
                    // Start, Alternate
                    formerWaveData = TrimWaveForDisplay(waveData, div, begin, right, true, true);
                    latterWaveData = TrimWaveForDisplay(waveData, div, begin, Width - left, true, true);
                }
                else if (Position == Position.End)
                {
                    // End, Alternate
                    formerWaveData = TrimWaveForDisplay(waveData, div, end, right, false, false);
                    latterWaveData = TrimWaveForDisplay(waveData, div, end, Width - left, false, false);
                }
            }

            var formerWavePoints = new Point[formerWaveData.Length];
            var latterWavePoints = new Point[latterWaveData.Length];

            for (int i = 0; i < formerWavePoints.Length; i++)
            {
                formerWavePoints[i] = new Point(right - 1 - i, (int)(-formerWaveData[i] * height / 2 * scale + height / 2));
            }
            for (int i = 0; i < latterWavePoints.Length; i++)
            {
                latterWavePoints[i] = new Point(left + i, (int)(-latterWaveData[i] * height / 2 * scale + height / 2));
            }

            using var waveformPen = new Pen(options.Color.Waveform, 1);

            if (formerWavePoints.Length > 1)
            {
                g.DrawLines(waveformPen, formerWavePoints);
            }
            if (latterWavePoints.Length > 1)
            {
                g.DrawLines(waveformPen, latterWavePoints);
            }
        }

        private float[] TrimWaveForDisplay(float[] buffer, double scaleReciprocal, int trimBegin, int samples, bool isLeft, bool isForward)
        {
            if (scaleReciprocal <= 0)
            {
                scaleReciprocal = 1;
            }

            if (trimBegin < 0)
            {
                trimBegin = 0;
            }

            int startPos = 0;

            // Left
            if (isLeft)
            {
                if (trimBegin > buffer.Length - 1)
                {
                    startPos = buffer.Length;
                }
                else
                {
                    startPos = trimBegin;
                }
            }
            else
            {
                // Right
                //startPos = buffer.Length - trimBegin - (int)(scaleReciprocal * samples);
                startPos = trimBegin - (int)(scaleReciprocal * samples);
                //if (startPos < 0)
                //{
                //    startPos = 0;
                //}
            }

            // Adjust result wave length.
            int trimEnd = startPos + (int)(scaleReciprocal * samples);
            if (trimEnd > buffer.Length)
            {
                samples = (int)((buffer.Length - startPos) / scaleReciprocal);
            }

            if (samples < 0)
            {
                samples = 0;
            }

            var result = new float[samples];

            if (isForward)
            {
                for (int i = 0; i < samples; i++)
                {
                    int pos = (int)(i * scaleReciprocal) + startPos;
                    if (pos >= 0 && pos < buffer.Length)
                    {
                        result[i] = buffer[pos];
                    }
                }
            }
            else
            {
                for (int i = 0; i < samples; i++)
                {
                    int pos = (int)(i * scaleReciprocal) + startPos;
                    if (pos >= 0 && pos < buffer.Length)
                    {
                        result[samples - 1 - i] = buffer[pos];
                    }
                }
            }

            return result;
        }

        private List<PointF> ComputeWaveformPoints()
        {
            var width = Size.Width;
            var bufferLength = buffer.Length;
            var step = bufferLength / width;
            var minMaxPairs = new MinMaxPair[width];
            List<PointF> result = new List<PointF>();

            if (buffer.Length == 0)
            {
                return result;
            }
            else if (width > bufferLength)
            {
                var amplitudeMax = buffer.Max(value => Math.Abs(value));
                var factor = 1 / amplitudeMax * options.WaveformHeightRate * Height / 2;

                for (int i = 0; i < bufferLength; i++)
                {
                    var x = (float)i * width / bufferLength;
                    var y = buffer[i] * factor;
                    result.Add(new PointF(x, y));
                }
                return result;
            }
            else
            {
                for (int i = 0; i < width; i++)
                {
                    var max = float.MinValue;
                    var min = float.MaxValue;

                    for (int j = 0; j < step; j++)
                    {
                        var index = i * step + j;
                        max = Math.Max(max, buffer[i * step + j]);
                        min = Math.Min(max, buffer[i * step + j]);
                    }

                    minMaxPairs[i] = new MinMaxPair() { Min = min, Max = max };
                }

                var amplitudeMax = minMaxPairs.Max(pair => Math.Max(Math.Abs(pair.Min), Math.Abs(pair.Max)));
                var factor = 1 / amplitudeMax * options.WaveformHeightRate * Height / 2;

                for (int i = 0; i < minMaxPairs.Length; i++)
                {
                    var min = minMaxPairs[i].Min;
                    var max = minMaxPairs[i].Max;
                }

                for (int i = 0; i < minMaxPairs.Length; i++)
                {
                    var min = minMaxPairs[i].Min;
                    var max = minMaxPairs[i].Max;
                    result.Add(new PointF(i, min * factor + Height / 2));
                    result.Add(new PointF(i, max * factor + Height / 2));
                }

                return result;
            }
        }
    }
}
