﻿#nullable enable

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace AudioLoopMaker
{
    class PictureBoxWaveSummary : PictureBox
    {
        WaveViewOptions options = new WaveViewOptions();


        Point? sectionBeginMoveBeginLocation = null;
        int startX;
        int startXdelta;

        Point? sectionEndMoveBeginLocation = null;
        int endX;
        int endXdelta;

        int cursorX;

        float[] buffer = new float[0];
        int startPosition;
        int endPosition;
        int cursorPosition;

        PointF[] waveformPoints = new PointF[0];
        PointF[] waveformDarkPoints = new PointF[0];

        public event EventHandler? SectionStartChanged;
        public event EventHandler? SectionEndChanged;

        internal WaveViewOptions WaveViewOptions { get => options; set => options = value; }

        public float[] Buffer
        {
            get => buffer;
            set
            {
                buffer = value;
                StartPosition = 0;
                EndPosition = buffer.Length;
                waveformPoints = ComputeWaveformPoints().ToArray();
                ComputeWaveformDarkPoints();
                Invalidate();
            }
        }

        public int StartPosition
        {
            get => startPosition;
            set
            {
                startPosition = value;
                startPosition = startPosition > 0 ? startPosition : 0;
                startPosition = startPosition < endPosition ? startPosition : endPosition;
                if (buffer.Length > 0)
                {
                    startX = (int)((double)startPosition * Width / buffer.Length);
                }
                Invalidate();
            }
        }

        public int EndPosition
        {
            get => endPosition;
            set
            {
                endPosition = value;
                endPosition = endPosition > 0 ? endPosition : 0;
                endPosition = endPosition < startPosition ? startPosition : endPosition;
                if (buffer.Length > 0)
                {
                    endX = (int)((double)endPosition * Width / buffer.Length);
                }
                Invalidate();
            }
        }

        public int CursorPosition
        {
            get => cursorPosition;
            set
            {
                cursorPosition = value;
                cursorPosition = cursorPosition > 0 ? cursorPosition : 0;
                if (buffer.Length > 0)
                {
                    cursorX = (int)((double)cursorPosition * Width / buffer.Length);
                }
                Invalidate();
            }
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            waveformPoints = ComputeWaveformPoints().ToArray();
            ComputeWaveformDarkPoints();
            if (buffer.Length > 0)
            {
                startX = (int)((double)startPosition * Width / buffer.Length);
                endX = (int)((double)endPosition * Width / buffer.Length);
                cursorX = (int)((double)cursorPosition * Width / buffer.Length);
            }
            Invalidate();
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if (e.X >= startX && e.X < startX + options.GripSize.Width
                && e.Y >= 0 && e.Y < options.GripSize.Height)
            {
                sectionBeginMoveBeginLocation = e.Location;
                startXdelta = startX - e.X;
            }
            else if (e.X >= endX - options.GripSize.Width
                && e.X <= endX && e.Y >= Height - options.GripSize.Height && e.Y < Height)
            {
                sectionEndMoveBeginLocation = e.Location;
                endXdelta = endX - e.X;
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            var gripSize = options.GripSize;
            if (sectionBeginMoveBeginLocation != null)
            {
                var x = e.X + startXdelta;
                if (x >= 0 && x <= endX)
                {
                    startX = x;
                    startPosition = (int)((double)startX * buffer.Length / Width);
                    SectionStartChanged?.Invoke(this, EventArgs.Empty);
                    Invalidate();
                    return;
                }
            }
            else if (sectionEndMoveBeginLocation != null)
            {
                var x = e.X + endXdelta;
                if (x >= startX && x < Width)
                {
                    endX = x;
                    if (endX < Width - 1)
                    {
                        endPosition = (int)((double)(endX + 1) * buffer.Length / Width);
                    }
                    else
                    {
                        endPosition = buffer.Length;
                    }
                    SectionEndChanged?.Invoke(this, EventArgs.Empty);
                    Invalidate();
                    return;
                }
            }
            else
            {
                if (e.X >= startX && e.X < startX + gripSize.Width && e.Y >= 0 && e.Y < gripSize.Height)
                {
                    Cursor = Cursors.Hand;
                }
                else if (e.X >= endX - gripSize.Width && e.X <= endX && e.Y >= Height - gripSize.Height && e.Y < Height)
                {
                    Cursor = Cursors.Hand;
                }
                else
                {
                    Cursor = Cursors.Default;
                }
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            Cursor = Cursors.Default;

            if (sectionBeginMoveBeginLocation != null)
            {
                SectionStartChanged?.Invoke(this, EventArgs.Empty);
            }
            else if (sectionEndMoveBeginLocation != null)
            {
                SectionEndChanged?.Invoke(this, EventArgs.Empty);
            }

            sectionBeginMoveBeginLocation = null;
            sectionEndMoveBeginLocation = null;
        }

        protected override void OnPaint(PaintEventArgs pe)
        {
            base.OnPaint(pe);

            var graphics = pe.Graphics;

            // Background
            graphics.Clear(options.Color.Background);

            // Waveform
            if (waveformPoints.Length > 1)
            {
                using var waveformPen = new Pen(options.Color.Waveform, 1);
                graphics.DrawLines(waveformPen, waveformPoints);
            }

            if (waveformDarkPoints.Length > 1)
            {
                using var waveformPen = new Pen(options.Color.WaveformDark, 1);
                graphics.DrawLines(waveformPen, waveformDarkPoints);
            }

            // Cursor
            using var cursorPen = new Pen(options.Color.Cursor, 1);
            graphics.DrawLine(cursorPen, cursorX, 0, cursorX, Height);

            // Section annotations
            using var sectionStartBrush = new SolidBrush(options.Color.SectionStart);
            using var sectionStartPen = new Pen(options.Color.SectionStart, 1);
            using var sectionEndBrush = new SolidBrush(options.Color.SectionEnd);
            using var sectionEndPen = new Pen(options.Color.SectionEnd, 1);

            var gw = options.GripSize.Width;
            var gh = options.GripSize.Height;
            graphics.FillRectangle(sectionStartBrush, startX, 0, gw, gh);
            graphics.DrawLine(sectionStartPen, startX, 0, startX, Height);

            graphics.FillRectangle(sectionEndBrush, endX - gw, Height - gh, gw, gh);
            graphics.DrawLine(sectionEndPen, endX, 0, endX, Height);
        }

        private List<PointF> ComputeWaveformPoints()
        {
            var width = Size.Width;
            var bufferLength = buffer.Length;
            var step = bufferLength / width;
            var minMaxPairs = new MinMaxPair[width];
            List<PointF> result = new List<PointF>();

            if (buffer.Length == 0)
            {
                return result;
            }
            else if (width > bufferLength)
            {
                var amplitudeMax = buffer.Max(value => Math.Abs(value));
                var factor = 1 / amplitudeMax * options.WaveformHeightRate * Height / 2;

                for (int i = 0; i < bufferLength; i++)
                {
                    var x = (float)i * width / (bufferLength - 1);
                    var y = -buffer[i] * factor + Height / 2;
                    result.Add(new PointF(x, y));
                }
                return result;
            }
            else
            {
                for (int i = 0; i < width; i++)
                {
                    var max = float.MinValue;
                    var min = float.MaxValue;

                    for (int j = 0; j < step; j++)
                    {
                        var index = i * step + j;
                        max = Math.Max(max, buffer[i * step + j]);
                        min = Math.Min(min, buffer[i * step + j]);
                    }

                    minMaxPairs[i] = new MinMaxPair() { Min = min, Max = max };
                }

                var amplitudeMax = minMaxPairs.Max(pair => Math.Max(Math.Abs(pair.Min), Math.Abs(pair.Max)));
                var factor = 1 / amplitudeMax * options.WaveformHeightRate * Height / 2;

                for (int i = 0; i < minMaxPairs.Length; i++)
                {
                    var min = minMaxPairs[i].Min;
                    var max = minMaxPairs[i].Max;
                    result.Add(new PointF(i, -min * factor + Height / 2));
                    result.Add(new PointF(i, -max * factor + Height / 2));
                }

                return result;
            }
        }
        
        private void ComputeWaveformDarkPoints()
        {
            var length = waveformPoints.Length;
            waveformDarkPoints = new PointF[waveformPoints.Length];
            for (int i = 0; i < length; i++)
            {
                var x = waveformPoints[i].X;
                var y = (waveformPoints[i].Y - Height / 2) * 0.5f + Height / 2;
                waveformDarkPoints[i] = new PointF(x, y);
            }
        }
    }
}
