﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Bvews.Wave;

namespace Bvews.Wave
{
    public class WaveDataReader
    {
        long positionCurrent;
        long positionDataBegin;
        long positionDataEnd;

        int blockAlign;
        int blockCount;

        WaveFormat waveFormat;
        WaveInfo waveInfo;

        FileStream fileStream;
        BinaryReader binaryReader;

        public WaveDataReader(string path)
        {
            waveInfo = new WaveInfo(path);
            waveFormat = waveInfo.WaveFormat;
            if (waveInfo.IsLoadSuccessful)
            {
                fileStream = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.Read);
                binaryReader = new BinaryReader(fileStream);

                positionDataBegin = waveInfo.ChunkData.Position;
                positionDataEnd = waveInfo.ChunkData.Position + waveInfo.ChunkData.Size;
                blockAlign = waveInfo.WaveFormat.BlockAlign;
                blockCount = (int)waveInfo.ChunkData.Size / waveInfo.WaveFormat.BlockAlign;
            }
        }

        public void Dispose()
        {
            Close();

            if (binaryReader != null)
            {
                //binaryReader.Dispose();
            }
            if (fileStream != null)
            {
                fileStream.Dispose();
            }
        }

        private void Close()
        {
            if (binaryReader != null)
            {
                binaryReader.Close();
            }
            if (fileStream != null)
            {
                fileStream.Close();
            }
        }

        public WaveData Read()
        {
            WaveData result = new WaveData();
            result.WaveInfo = waveInfo;
            result.Buffers = ReadSamplesAsFloat();
            return result;
        }

        public float[][] ReadSamplesAsFloat()
        {
            fileStream.Position = positionDataBegin;

            var channels = waveInfo.WaveFormat.Channels;
            var offset = GetOffset(waveFormat);
            var result = new float[channels][];
            var sampleAlign = blockAlign / channels;

            for (int i = 0; i < channels; i++)
            {
                result[i] = new float[blockCount];
            }

            for (int i = 0; i < blockCount; i++)
            {
                for (int j = 0; j < channels; j++)
                {
                    try
                    {
                        var bytes = binaryReader.ReadBytes(sampleAlign);
                        result[j][i] = ReadAsFloat(bytes, waveFormat, 0);
                    }
                    catch
                    {
                        Console.WriteLine(fileStream.Position);
                        throw;
                    }
                }
            }

            return result;
        }

        private float ReadAsFloat(byte[] bytes, WaveFormat format, int offset)
        {
            switch (format.FormatTag)
            {
                case WaveFormatTag.Pcm:
                    switch (format.BitsPerSample)
                    {
                        case 8:
                            return (bytes[offset] - 128) / 128f;
                        case 16:
                            return (BitConverter.ToInt16(bytes, offset) / 32768f);
                        case 24:
                            var value = bytes[0] | BitConverter.ToInt16(bytes, 1) << 8;
                            return (value / 8388608f);
                        case 32:
                            return (BitConverter.ToInt32(bytes, offset) / 2147483648f);
                    }
                    break;
                case WaveFormatTag.IeeeFloat:
                    switch (format.BitsPerSample)
                    {
                        case 32:
                            return BitConverter.ToSingle(bytes, offset);
                        case 64:
                            return (float)BitConverter.ToDouble(bytes, offset);
                    }
                    break;
            }

            return 0.0f;
        }

        private static short ConvertToShort(byte[] bytes, WaveFormat format, int offset)
        {
            switch (format.FormatTag)
            {
                case WaveFormatTag.Pcm:
                    switch (format.BitsPerSample)
                    {
                        case 8:
                            return (short)((bytes[offset] - 128) << 8);
                        case 16:
                            return BitConverter.ToInt16(bytes, offset);
                        case 24:
                            return BitConverter.ToInt16(bytes, offset + 1);
                        case 32:
                            return BitConverter.ToInt16(bytes, offset + 2);
                    }
                    break;
                case WaveFormatTag.IeeeFloat:
                    switch (format.BitsPerSample)
                    {
                        case 32:
                            return (short)(BitConverter.ToSingle(bytes, offset) * 32768);
                        //return (short)(BitConverter.ToSingle(bytes, offset) * 32768 - 0.5);
                        case 64:
                            return (short)(BitConverter.ToDouble(bytes, offset) * 32768);
                        //return (short)(BitConverter.ToDouble(bytes, offset) * 32768 - 0.5);
                    }
                    break;
            }

            return 0;
        }

        private static byte[] ConvertToBytes(float value, WaveFormat format, int offset)
        {
            switch (format.FormatTag)
            {
                case WaveFormatTag.Pcm:
                    switch (format.BitsPerSample)
                    {
                        case 8:
                            return new byte[] { (byte)(value * 128) };
                        case 16:
                            return BitConverter.GetBytes((short)(value * 32768));
                        case 24:
                            var bytes = BitConverter.GetBytes((int)(value * 8388608));
                            Array.Resize(ref bytes, 3);
                            return bytes;
                        case 32:
                            return BitConverter.GetBytes((int)(value * 2147483648));
                    }
                    break;
                case WaveFormatTag.IeeeFloat:
                    switch (format.BitsPerSample)
                    {
                        case 32:
                            return BitConverter.GetBytes(value);
                        case 64:
                            return BitConverter.GetBytes((double)value);
                    }
                    break;
            }

            return new byte[0];
        }

        private static byte[] ConvertToBytes(short value, WaveFormat format, int offset)
        {
            switch (format.FormatTag)
            {
                case WaveFormatTag.Pcm:
                    switch (format.BitsPerSample)
                    {
                        case 8:
                            return new byte[] { (byte)(value >> 8) };
                        case 16:
                            return BitConverter.GetBytes(value);
                        case 24:
                            var bytes = BitConverter.GetBytes((int)(value << 8));
                            Array.Resize(ref bytes, 3);
                            return bytes;
                        case 32:
                            return BitConverter.GetBytes((int)(value << 16));
                    }
                    break;
                case WaveFormatTag.IeeeFloat:
                    switch (format.BitsPerSample)
                    {
                        case 32:
                            return BitConverter.GetBytes(value / 32768.0f);
                        case 64:
                            return BitConverter.GetBytes(value / 32768.0);
                    }
                    break;
            }

            return new byte[0];
        }

        private static int GetOffset(WaveFormat format)
        {
            if (format.Channels != 2)
            {
                return 0;
            }
            else
            {
                return format.BitsPerSample / 8 / 2;
            }
        }
    }

}
