﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Bvews.Wave;

namespace Bvews.Wave
{
    public class WaveDataWriter
    {
        public void Write(string path, WaveData waveData)
        {
            Write(path, waveData.WaveInfo, waveData.Buffers);
        }

        public void Write(Stream stream, WaveData waveData)
        {
            Write(stream, waveData.WaveInfo, waveData.Buffers);
        }

        public void Write(string path, WaveInfo info, float[][] buffers)
        {
            using (var fs = new FileStream(path, FileMode.Create, FileAccess.Write))
            {
                Write(fs, info, buffers);
            }
        }

        public void Write(Stream stream, WaveInfo info, float[][] buffers)
        {
            BinaryWriter bw;
            try
            {
                bw = WriteWithBinaryWriter(stream, info, buffers);
                bw.Flush();
                bw.Close();
            }
            catch
            {
                throw;
            }
            finally
            {
            }
        }

        public BinaryWriter WriteWithBinaryWriter(Stream stream, WaveInfo info, float[][] buffers)
        {
            var bw = new BinaryWriter(stream);

            var format = info.WaveFormat;
            var channels = format.Channels;
            format.Channels = (short)channels;

            var length = buffers.Min(s => s.Length);

            var dataSize = buffers.Length * length * format.BitsPerSample / 8;
            var riffSize = dataSize + 36;

            bw.Write(info.ChunkRiff.Id);
            bw.Write(riffSize);
            bw.Write(info.ChunkRiff.FormType);

            bw.Write(info.ChunkFmt.Id);
            //bw.Write(info.ChunkFmt.Size);
            // Fixed "format chunk" fftSize.
            bw.Write(0x0010);
            bw.Write(format.GenerateBytes());

            bw.Write(info.ChunkData.Id);
            bw.Write(dataSize);

            for (int i = 0; i < length; i++)
            {
                for (int j = 0; j < channels; j++)
                {
                    switch (format.FormatTag)
                    {
                        case WaveFormatTag.Pcm:
                            switch (format.BitsPerSample)
                            {
                                case 8:
                                    WriteSampleAs8BitPcm(bw, buffers[j][i]);
                                    break;
                                case 16:
                                    WriteSampleAs16BitPcm(bw, buffers[j][i]);
                                    break;
                                case 24:
                                    WriteSampleAs24BitPcm(bw, buffers[j][i]);
                                    break;
                                case 32:
                                    WriteSampleAs32BitPcm(bw, buffers[j][i]);
                                    break;
                            }
                            break;
                        case WaveFormatTag.IeeeFloat:
                            switch (format.BitsPerSample)
                            {
                                case 32:
                                    WriteSampleAsFloat(bw, buffers[j][i]);
                                    break;
                                case 64:
                                    WriteSampleAsDouble(bw, buffers[j][i]);
                                    break;
                            }
                            break;
                    }
                }
            }

            return bw;
        }

        private void WriteSampleAs8BitPcm(BinaryWriter bw, float sample)
        {
            var value = (int)(128 * sample + 128 + 0.5);
            if (value < 0)
            {
                bw.Write((byte)0);
            }
            else if (value > 255)
            {
                bw.Write((byte)255);
            }
            else
            {
                bw.Write((byte)value);
            }
        }

        private void WriteSampleAs16BitPcm(BinaryWriter bw, float sample)
        {
            var value = (int)(32768 * sample + 0.5);
            if (value < -32768)
            {
                bw.Write((short)-32768);
            }
            else if (value > 32767)
            {
                bw.Write((short)32767);
            }
            else
            {
                bw.Write((short)value);
            }
        }

        private void WriteSampleAs24BitPcm(BinaryWriter bw, float sample)
        {
            var value = (int)(8388608 * sample + 0.5);
            if (value < -8388608)
            {
                Write(bw, -8388608);
            }
            else if (value > 8388607)
            {
                Write(bw, 8388607);
            }
            else
            {
                Write(bw, value);
            }
        }

        private void Write(BinaryWriter bw, int sample)
        {
            byte @byte = (byte)(sample & 0xFF);
            short @short = (short)(sample >> 8);
            bw.Write(@byte);
            bw.Write(@short);
        }

        private void WriteSampleAs32BitPcm(BinaryWriter bw, float sample)
        {
            var value = (int)(2147483648 * sample + 0.5);
            if (value < -2147483648)
            {
                bw.Write(-2147483648);
            }
            else if (value > 2147483647)
            {
                bw.Write(2147483647);
            }
            else
            {
                bw.Write(value);
            }
        }

        private void WriteSampleAsFloat(BinaryWriter bw, float sample)
        {
            bw.Write(sample);
        }

        private void WriteSampleAsDouble(BinaryWriter bw, float sample)
        {
            bw.Write((double)sample);
        }
    }

}
